"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var GunEvent_1 = require("./ControlFlow/GunEvent");
var ChainGunLink = /** @class */ (function () {
    function ChainGunLink(chain, key, parent) {
        this.key = key;
        if (!parent)
            this.soul = key;
        this._chain = chain;
        this._parent = parent;
        this._hasReceived = false;
        this._updateEvent = new GunEvent_1.GunEvent(this.getPath().join('|'));
    }
    ChainGunLink.prototype.getPath = function () {
        if (this._parent)
            return __spreadArrays(this._parent.getPath(), [this.key]);
        return [this.key];
    };
    /**
     * Traverse a location in the graph
     *
     * @param key Key to read data from
     * @param cb
     * @returns New chain context corresponding to given key
     */
    ChainGunLink.prototype.get = function (key, cb) {
        return new this.constructor(this._chain, key, this);
    };
    /**
     * Move up to the parent context on the chain.
     *
     * Every time a new chain is created, a reference to the old context is kept to go back to.
     *
     * @param amount The number of times you want to go back up the chain. {-1} or {Infinity} will take you to the root.
     * @returns a parent chain context
     */
    ChainGunLink.prototype.back = function (amount) {
        if (amount === void 0) { amount = 1; }
        if (amount < 0 || amount === Infinity)
            return this._chain;
        if (amount === 1)
            return this._parent || this._chain;
        return this.back(amount - 1);
    };
    /**
     * Save data into gun, syncing it with your connected peers.
     *
     * You do not need to re-save the entire object every time, gun will automatically
     * merge your data into what already exists as a "partial" update.
     *
     * @param value the data to save
     * @param cb an optional callback, invoked on each acknowledgment
     * @returns same chain context
     */
    ChainGunLink.prototype.put = function (value, cb) {
        this._chain.graph.putPath(this.getPath(), value, cb);
        return this;
    };
    /**
     * Add a unique item to an unordered list.
     *
     * Works like a mathematical set, where each item in the list is unique.
     * If the item is added twice, it will be merged.
     * This means only objects, for now, are supported.
     *
     * @param data should be a gun reference or an object
     * @param cb The callback is invoked exactly the same as .put
     * @returns chain context for added object
     */
    ChainGunLink.prototype.set = function (data, cb) {
        var _a, _b;
        if (data instanceof ChainGunLink && data.soul) {
            this.put((_a = {},
                _a[data.soul] = {
                    '#': data.soul
                },
                _a), cb);
        }
        else if (data && data._ && data._['#']) {
            this.put((_b = {},
                _b[data && data._ && data._['#']] = data,
                _b), cb);
        }
        else {
            throw new Error('set() is only partially supported');
        }
    };
    /**
     * Register a callback for when it appears a record does not exist
     *
     * If you need to know whether a property or key exists, you can check with .not.
     * It will consult the connected peers and invoke the callback if there's reasonable certainty that none of them have the data available.
     *
     * @param cb If there's reason to believe the data doesn't exist, the callback will be invoked. This can be used as a check to prevent implicitly writing data
     * @returns same chain context
     */
    ChainGunLink.prototype.not = function (cb) {
        var _this = this;
        this.promise().then(function (val) { return typeof val === 'undefined' && cb(_this.key); });
        return this;
    };
    /**
     * Change the configuration of this chain link
     *
     * @param options
     * @returns same chain context
     */
    ChainGunLink.prototype.opt = function (options) {
        return this;
    };
    /**
     * Get the current data without subscribing to updates. Or undefined if it cannot be found.
     *
     * @param cb The data is the value for that chain at that given point in time. And the key is the last property name or ID of the node.
     * @returns same chain context
     */
    ChainGunLink.prototype.once = function (cb) {
        var _this = this;
        this.promise().then(function (val) { return cb(val, _this.key); });
        return this;
    };
    /**
     * Subscribe to updates and changes on a node or property in realtime.
     *
     * Triggered once initially and whenever the property or node you're focused on changes,
     * Since gun streams data, the callback will probably be called multiple times as new chunk comes in.
     *
     * To remove a listener call .off() on the same property or node.
     *
     * @param cb The callback is immediately fired with the data as it is at that point in time.
     * @returns same chain context
     */
    ChainGunLink.prototype.on = function (cb) {
        this._updateEvent.on(cb);
        if (!this._endQuery) {
            this._endQuery = this._chain.graph.query(this.getPath(), this._onQueryResponse.bind(this));
        }
        if (this._hasReceived)
            cb(this._lastValue, this.key);
        return this;
    };
    /**
     * Unsubscribe one or all listeners subscribed with on
     *
     * @returns same chain context
     */
    ChainGunLink.prototype.off = function (cb) {
        if (cb) {
            this._updateEvent.off(cb);
            if (this._endQuery && !this._updateEvent.listenerCount()) {
                this._endQuery();
            }
        }
        else {
            if (this._endQuery) {
                this._endQuery();
            }
            this._updateEvent.reset();
        }
        return this;
    };
    ChainGunLink.prototype.promise = function (opts) {
        var _this = this;
        if (opts === void 0) { opts = { timeout: 0 }; }
        return new Promise(function (ok) {
            var cb = function (val) {
                ok(val);
                _this.off(cb);
            };
            _this.on(cb);
            if (opts.timeout)
                setTimeout(function () { return ok(); }, opts.timeout);
        });
    };
    ChainGunLink.prototype.then = function (fn) {
        return this.promise().then(fn);
    };
    /**
     * Iterates over each property and item on a node, passing it down the chain
     *
     * Not yet supported
     *
     * Behaves like a forEach on your data.
     * It also subscribes to every item as well and listens for newly inserted items.
     *
     * @returns a new chain context holding many chains simultaneously.
     */
    ChainGunLink.prototype.map = function () {
        throw new Error("map() isn't supported yet");
        return this.get(''); // Special case key
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.path = function (path) {
        throw new Error('No plans to support this');
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.open = function (cb) {
        throw new Error('No plans to support this');
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.load = function (cb) {
        throw new Error('No plans to support this');
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.bye = function () {
        throw new Error('No plans to support this');
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.later = function () {
        throw new Error('No plans to support this');
    };
    /**
     * No plans to support this
     */
    ChainGunLink.prototype.unset = function (node) {
        throw new Error('No plans to support this');
    };
    ChainGunLink.prototype._onQueryResponse = function (value) {
        this._updateEvent.trigger(value, this.key);
        this._lastValue = value;
        this._hasReceived = true;
    };
    return ChainGunLink;
}());
exports.ChainGunLink = ChainGunLink;
//# sourceMappingURL=ChainGunLink.js.map